import abc
from inspect import signature
from typing import Dict, Optional, List

from langcodes import closest_match
from ovos_bus_client.message import Message
from ovos_bus_client.message import dig_for_message
from ovos_config.config import Configuration
from ovos_utils.lang import standardize_lang_tag
from ovos_utils.log import LOG
from ovos_utils.skills import get_non_properties
from padacioso import IntentContainer

from ovos_workshop.decorators.killable import AbortEvent, killable_event, AbortQuestion
from ovos_workshop.resource_files import ResourceFile
from ovos_workshop.skills.ovos import OVOSSkill


class ConversationalSkill(OVOSSkill):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.converse_matchers = {}

    def activate(self, duration_minutes=None):
        """
        Mark this skill as active and push to the top of the active skills list.
        This enables converse method to be called even without skill being
        used in last 5 minutes.

        :param duration_minutes: duration in minutes for skill to remain active
         (-1 for infinite)
        """
        if duration_minutes is None:
            duration_minutes = Configuration().get("converse", {}).get("timeout", 300) / 60  # convert to minutes

        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id

        m1 = msg.forward("intent.service.skills.activate",
                         data={"skill_id": self.skill_id,
                               "timeout": duration_minutes})
        self.bus.emit(m1)

    def deactivate(self):
        """
        Mark this skill as inactive and remove from the active skills list.
        This stops converse method from being called.
        """
        msg = dig_for_message() or Message("")
        if "skill_id" not in msg.context:
            msg.context["skill_id"] = self.skill_id
        self.bus.emit(msg.forward("intent.service.skills.deactivate",
                                  data={"skill_id": self.skill_id}))

    def _register_system_event_handlers(self):
        super()._register_system_event_handlers()
        self.add_event(f"{self.skill_id}.converse.ping", self._handle_converse_ack, speak_errors=False)
        self.add_event(f"{self.skill_id}.converse.request", self._handle_converse_request, speak_errors=False)
        self.add_event(f"{self.skill_id}.activate", self.handle_activate, speak_errors=False)
        self.add_event(f"{self.skill_id}.deactivate", self.handle_deactivate, speak_errors=False)
        self.add_event("intent.service.skills.deactivated", self._handle_skill_deactivated, speak_errors=False)
        self.add_event("intent.service.skills.activated", self._handle_skill_activated, speak_errors=False)

    def _register_decorated(self):
        """
        Register all intent handlers that are decorated with an intent.

        Looks for all functions that have been marked by a decorator
        and read the intent data from them.  The intent handlers aren't the
        only decorators used.  Skip properties as calling getattr on them
        executes the code which may have unintended side effects
        """
        super()._register_decorated()
        for attr_name in get_non_properties(self):
            method = getattr(self, attr_name)

            if hasattr(method, 'converse_intents'):
                for intent_file in getattr(method, 'converse_intents'):
                    self.register_converse_intent(intent_file, method)

    def register_converse_intent(self, intent_file, handler):
        """ converse padacioso intents """
        name = f'{self.skill_id}.converse:{intent_file}'
        fuzzy = not self.settings.get("strict_intents", False)

        for lang in self.native_langs:
            self.converse_matchers[lang] = IntentContainer(fuzz=fuzzy)

            resources = self.load_lang(self.res_dir, lang)
            resource_file = ResourceFile(resources.types.intent, intent_file)
            if resource_file.file_path is None:
                self.log.error(f'Unable to find "{intent_file}"')
                continue
            filename = str(resource_file.file_path)

            with open(filename) as f:
                samples = [l.strip() for l in f.read().split("\n")
                           if l and not l.startswith("#")]

            self.converse_matchers[lang].add_intent(name, samples)

        self.add_event(name, handler, 'mycroft.skill.handler')

    def _get_closest_lang(self, lang: str) -> Optional[str]:
        if self.converse_matchers:
            lang = standardize_lang_tag(lang)
            closest, score = closest_match(lang, list(self.converse_matchers.keys()))
            # https://langcodes-hickford.readthedocs.io/en/sphinx/index.html#distance-values
            # 0 -> These codes represent the same language, possibly after filling in values and normalizing.
            # 1- 3 -> These codes indicate a minor regional difference.
            # 4 - 10 -> These codes indicate a significant but unproblematic regional difference.
            if score < 10:
                return closest
        return None

    def _handle_converse_ack(self, message: Message):
        """
        Inform skills service if we want to handle converse. Individual skills
        must implement self.can_answer
        @param message: `{self.skill_id}.converse.ping` Message
        """
        self.bus.emit(message.reply(
            "skill.converse.pong",
            data={"skill_id": self.skill_id,
                  "can_handle": self.can_answer(message)},
            context={"skill_id": self.skill_id}))

    def _handle_skill_activated(self, message: Message):
        """
        Intent service activated a skill. If it was this skill,
        emit a skill activation message.
        @param message: `intent.service.skills.activated` Message
        """
        if message.data.get("skill_id") == self.skill_id:
            self.bus.emit(message.forward(f"{self.skill_id}.activate"))

    def _handle_skill_deactivated(self, message):
        """
        Intent service deactivated a skill. If it was this skill,
        emit a skill deactivation message.
        @param message: `intent.service.skills.deactivated` Message
        """
        if message.data.get("skill_id") == self.skill_id:
            self.bus.emit(message.forward(f"{self.skill_id}.deactivate"))

    def _on_timeout(self):
        """_handle_converse_request timed out and was forcefully killed by ovos-core"""
        message = dig_for_message()
        self.bus.emit(message.forward(
            f"{self.skill_id}.converse.killed",
            data={"error": "timed out"}))

    @killable_event("ovos.skills.converse.force_timeout",
                    callback=_on_timeout, check_skill_id=True)
    def _handle_converse_request(self, message: Message):
        """
        If this skill is requested and supports converse, handle the user input
        with `converse`.
        @param message: `{self.skill_id}.converse.request` Message
        """
        # check if a conversational intent triggered
        # these are skill specific intents that may trigger instead of converse
        if self._handle_converse_intents(message):
            self.bus.emit(message.reply('skill.converse.response',
                                        {"skill_id": self.skill_id, "result": True}))
            return

        try:
            # converse can have multiple signatures
            params = signature(self.converse).parameters
            kwargs = {"message": message,
                      "utterances": message.data['utterances'],
                      "lang": standardize_lang_tag(message.data['lang'])}
            kwargs = {k: v for k, v in kwargs.items() if k in params}

            result = self.converse(**kwargs)

            self.bus.emit(message.reply('skill.converse.response',
                                        {"skill_id": self.skill_id,
                                         "result": result}))
        except (AbortQuestion, AbortEvent):
            self.bus.emit(message.reply('skill.converse.response',
                                        {"skill_id": self.skill_id,
                                         "error": "killed",
                                         "result": False}))
        except Exception as e:
            LOG.error(e)
            self.bus.emit(message.reply('skill.converse.response',
                                        {"skill_id": self.skill_id,
                                         "error": repr(e),
                                         "result": False}))

    def _handle_converse_intents(self, message):
        """ called before converse method
        this gives active skills a chance to parse their own intents and
        consume the utterance, see conversational_intent decorator for usage
        """
        lang = self._get_closest_lang(self.lang)
        if lang is None:  # no intents registered for this lang
            return None

        best_score = 0
        response = None

        for utt in message.data['utterances']:
            match = self.converse_matchers[lang].calc_intent(utt)
            if match.get("conf", 0) > best_score:
                best_score = match["conf"]
                response = message.forward(match["name"], match["entities"])

        if not response or best_score < self.settings.get("min_intent_conf", 0.5):
            return False

        # send intent event
        self.bus.emit(response)
        return True

    @abc.abstractmethod
    def can_answer(self, message: Message) -> bool:
        """
        Determine if the skill can handle the given utterance during the converse phase.

        Override this method to implement custom logic for assessing whether the skill
        is capable of answering the user's query based on the utterance and session context.

        Notes:
            - Utterance transcriptions are available via `message.data["utterances"]`.
            - The session (e.g., to access language) can be retrieved using:
              `session = SessionManager.get(message)`.

        Args:
            message (Message): The message containing user utterances and metadata.

        Returns:
            bool: True if the skill can handle the query during converse; False otherwise.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def converse(self, message: Message):
        """
        Handle the user's utterance if this skill was selected via `can_answer`.

        This method is called only if `can_answer` returned True and the skill was chosen
        to handle the user's input during a conversation.

        Args:
            message (Message): The message containing the user utterances to process.
        """
        raise NotImplementedError

    def handle_activate(self, message: Message):
        """
        Called when this skill is considered active by the intent service;
        converse method will be called with every utterance.
        Override this method to do any optional preparation.
        @param message: `{self.skill_id}.activate` Message
        """

    def handle_deactivate(self, message: Message):
        """
        Called when this skill is no longer considered active by the intent
        service; converse method will not be called until skill is active again.
        Override this method to do any optional cleanup.
        @param message: `{self.skill_id}.deactivate` Message
        """



import abc
from typing import Optional, Dict, Iterable

from ovos_bus_client.message import Message
from ovos_bus_client.util import get_message_lang
from ovos_utils.ocp import MediaType, MediaEntry, PlaybackType, Playlist, PlayerState
from ovos_utils.parse import match_one, MatchStrategy

from ovos_workshop.decorators import ocp_featured_media, ocp_search
from ovos_workshop.skills.common_play import OVOSCommonPlaybackSkill
from ovos_workshop.skills.ovos import _get_dialog


class OVOSGameSkill(OVOSCommonPlaybackSkill):
    """ To integrate with the OpenVoiceOS Common Playback framework

    "play" intent is shared with media and managed by OCP pipeline

    The class makes the skill available to queries from OCP
        - "skill_voc_filename" keyword argument is mandatory
          it defines the .voc file containing the keywords to match the game name

    bus events emitted to trigger this skill:
    - 'ovos.common_play.{self.skill_id}.play'
    - 'ovos.common_play.{self.skill_id}.pause'
    - 'ovos.common_play.{self.skill_id}.resume'
    - 'ovos.common_play.{self.skill_id}.stop'
    - 'ovos.common_play.{self.skill_id}.save'
    - 'ovos.common_play.{self.skill_id}.load'

    """

    def __init__(self, skill_voc_filename: str,
                 *args,
                 skill_icon: str = "",
                 game_image: str = "",
                 **kwargs):
        """IMPORTANT: contents of skill_voc_filename are crucial for intent matching
        without that ocp_pipeline might not recognize the skill as a game"""
        self.game_image = game_image
        super().__init__(skill_icon=skill_icon, skill_voc_filename=skill_voc_filename,
                         supported_media=[MediaType.GAME],
                         playback_handler=self.on_play_game,
                         pause_handler=self.on_pause_game,
                         resume_handler=self.on_resume_game,
                         *args, **kwargs)

    @ocp_featured_media()
    def _ocp_featured(self) -> Playlist:
        """ensure skill shows up in OCP GUI menu
        report the game as the only featured_track"""
        entry = MediaEntry(
            uri=f"skill:{self.skill_id}",
            title=self.skill_aliases[0],
            image=self.game_image,
            playback=PlaybackType.SKILL,
            media_type=MediaType.GAME,
            match_confidence=100,
            skill_icon=self.skill_icon
        )
        pl = Playlist(
            title=self.skill_aliases[0],
            image=self.game_image,
            playback=PlaybackType.SKILL,
            media_type=MediaType.GAME,
            match_confidence=100,
            skill_icon=self.skill_icon)
        pl.add_entry(entry)
        return pl

    @ocp_search()
    def _ocp_search(self, phrase: str, media_type: MediaType) -> Iterable[MediaEntry]:
        """match the game name when OCP is searching"""
        if media_type != MediaType.GAME:
            # only match if ocp_pipeline determined a game was wanted
            return

        _, score = match_one(phrase, self.skill_aliases,
                             strategy=MatchStrategy.DAMERAU_LEVENSHTEIN_SIMILARITY)
        conf = int(score * 100)
        if conf >= 50:
            entry = self._ocp_featured()[0]
            entry.match_confidence = conf
            yield entry

    @property
    def is_playing(self) -> bool:
        return self._playing.is_set()

    @property
    def is_paused(self) -> bool:
        return self._paused.is_set()

    @abc.abstractmethod
    def on_play_game(self):
        """called by ocp_pipeline when 'play XXX' matches the game"""

    @abc.abstractmethod
    def on_pause_game(self):
        """called by ocp_pipeline on 'pause' if game is being played"""

    @abc.abstractmethod
    def on_resume_game(self):
        """called by ocp_pipeline on 'resume/unpause' if game is being played and paused"""

    @abc.abstractmethod
    def on_stop_game(self):
        """called when game is stopped for any reason
        auto-save may be implemented here"""

    @abc.abstractmethod
    def on_save_game(self):
        """if your game has no save/load functionality you should
        speak a error dialog here"""

    @abc.abstractmethod
    def on_load_game(self):
        """if your game has no save/load functionality you should
        speak a error dialog here"""

    def stop_game(self):
        """to be called by skills if they want to stop game programatically"""
        if self.is_playing:
            self._paused.clear()
            self.gui.release()
            self.log.debug("changing OCP state: PlayerState.STOPPED ")
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))
            self._playing.clear()
            self.on_stop_game()
            return True
        return False

    def stop(self) -> bool:
        """NOTE: not meant to be called by the skill, this is a callback"""
        return self.stop_game()

    def calc_intent(self, utterance: str, lang: str, timeout=1.0) -> Optional[Dict[str, str]]:
        """helper to check what intent would be selected by ovos-core"""
        # let's see what intent ovos-core will assign to the utterance
        # NOTE: converse, common_query and fallbacks are not included in this check
        response = self.bus.wait_for_response(Message("intent.service.intent.get",
                                                      {"utterance": utterance, "lang": lang}),
                                              "intent.service.intent.reply",
                                              timeout=timeout)
        if not response:
            return None
        return response.data["intent"]


class ConversationalGameSkill(OVOSGameSkill):

    def on_save_game(self):
        """skills can override method to implement functioonality"""
        speech = _get_dialog("cant_save_game", self.lang)
        self.speak(speech)

    def on_load_game(self):
        """skills can override method to implement functioonality"""
        speech = _get_dialog("cant_load_game", self.lang)
        self.speak(speech)

    def on_pause_game(self):
        """called by ocp_pipeline on 'pause' if game is being played"""
        self._paused.set()
        self.acknowledge()
        # individual skills can change default value if desired
        if self.settings.get("pause_dialog", False):
            speech = _get_dialog("game_pause", self.lang)
            self.speak(speech)

    def on_resume_game(self):
        """called by ocp_pipeline on 'resume/unpause' if game is being played and paused"""
        self._paused.clear()
        self.acknowledge()
        # individual skills can change default value if desired
        if self.settings.get("pause_dialog", False):
            speech = _get_dialog("game_unpause", self.lang)
            self.speak(speech)

    @abc.abstractmethod
    def on_play_game(self):
        """called by ocp_pipeline when 'play XXX' matches the game"""

    @abc.abstractmethod
    def on_stop_game(self):
        """called when game is stopped for any reason
        auto-save may be implemented here"""

    @abc.abstractmethod
    def on_game_command(self, utterance: str, lang: str):
        """pipe user input that wasnt caught by intents to the game
        do any intent matching or normalization here
        don't forget to self.speak the game output too!
        """

    def on_abandon_game(self):
        """user abandoned game mid interaction

        auto-save is done before this method is called
        (if enabled in self.settings)

        on_game_stop will be called after this handler"""

    # converse
    def skill_will_trigger(self, utterance: str, lang: str, skill_id: Optional[str] = None, timeout=0.8) -> bool:
        """helper to check if this skill would be selected by ovos-core with the given utterance

        useful in converse method
            eg. return not self.will_trigger

        this example allows the utterance to be consumed via converse of using ovos-core intent parser
        ensuring it is always handled by the game skill regardless
        """
        # determine if an intent from this skill
        # will be selected by ovos-core
        id_to_check = skill_id or self.skill_id
        intent = self.calc_intent(utterance, lang, timeout=timeout)
        skill_id = intent["skill_id"] if intent else ""
        return skill_id == id_to_check

    @property
    def save_is_implemented(self) -> bool:
        """
        True if this skill implements a `save` method
        """
        return self.__class__.on_save_game is not ConversationalGameSkill.on_save_game

    def _autosave(self):
        """helper to save the game automatically if enabled in settings.json and implemented by skill"""
        if self.settings.get("auto_save", False) and self.save_is_implemented:
            self.on_save_game()

    def _async_cmd(self, message: Message):
        utterance = message.data["utterances"][0]
        lang = get_message_lang(message)
        self.log.debug(f"Piping utterance to game: {utterance}")
        self.on_game_command(utterance, lang)

    def converse(self, message: Message) -> bool:
        try:
            utterance = message.data["utterances"][0]
            lang = get_message_lang(message)
            # let the user implemented intents do the job if they can handle the utterance
            # otherwise pipe utterance to the game handler
            if self.skill_will_trigger(utterance, lang):
                self.log.debug("Skill intent will trigger, don't pipe utterance to game")
                return False

            if self.is_paused:
                self.log.debug("game is paused")
                # let ocp_pipeline unpause as appropriate
                return False

            try:
                self._autosave()
            except Exception as e:
                self.log.error(f"Autosave failed: {e}")

            if self.is_playing:
                # do this async so converse executes faster
                self.bus.once(f"{self.skill_id}.game_cmd", self._async_cmd)
                self.bus.emit(message.forward(f"{self.skill_id}.game_cmd", message.data))
                return True

            return False
        except (KeyError, IndexError) as e:
            self.log.error(f"Error processing game converse message: {e}")
            return False
        except Exception as e:
            self.log.exception(f"Unexpected error in game converse: {e}")
            return False

    def handle_deactivate(self, message: Message):
        """
        Called when this skill is no longer considered active by the intent service;
        means the user didn't interact with the game for a long time and intent parser will be released
        """
        try:
            if self.is_paused:
                self.log.info("Game is paused, keeping it active")
                self.activate()  # keep the game in active skills list so it can still converse
            elif self.is_playing:
                self._autosave()
                self.log.info("Game abandoned due to inactivity")
                self.on_abandon_game()
                self.stop_game()
        except Exception as e:
            self.log.exception(f"Error during game deactivation: {e}")

    def stop(self) -> bool:
        self._autosave()
        return super().stop()



from json_database import JsonStorageXDG


def settings2meta(settings, section_name="Skill Settings"):
    """ generates basic settingsmeta """
    fields = []

    for k, v in settings.items():
        if k.startswith("_"):
            continue
        label = k.replace("-", " ").replace("_", " ").title()
        if isinstance(v, bool):
            fields.append({
                "name": k,
                "type": "checkbox",
                "label": label,
                "value": str(v).lower()
            })
        if isinstance(v, str):
            fields.append({
                "name": k,
                "type": "text",
                "label": label,
                "value": v
            })
        if isinstance(v, int):
            fields.append({
                "name": k,
                "type": "number",
                "label": label,
                "value": str(v)
            })
    return {
        "skillMetadata": {
            "sections": [
                {
                    "name": section_name,
                    "fields": fields
                }
            ]
        }
    }


class PrivateSettings(JsonStorageXDG):
    def __init__(self, skill_id):
        super(PrivateSettings, self).__init__(skill_id)

    @property
    def settingsmeta(self):
        return settings2meta(self, self.name)
